package mngr

import (
	"encoding/json"
	"fmt"
	"github.com/hyperledger/fabric/common/util"
	"github.com/hyperledger/fabric/core/chaincode/shim"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/mock"
	"testing"
	"gitlab.smartblocklab.com/smartcontracts/nums"
	"gitlab.smartblocklab.com/smartcontracts/common/invoke"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincodetest"
)

func TestOfferManagerImpl_AddValidOffer(t *testing.T) {
	// Initialize offer manager
	mngr := offerManagerImpl{}

	testOperationID := "offerID1"

	offer := &OfferRequest{
		OperationID: testOperationID,
		OutUserInfo: OfferRequestInfo{
			UserID:       "outUserID",
			WalletID:     "outWalletID",
			AccountID:    "outAccountID",
			Amount:       nums.PlainCurrencyNumber{
				Amount: "100.0",
				CurrencyType: "BTC",
			},
		},
		InUserInfo: OfferRequestInfo{
			UserID:       "inUserID",
			WalletID:     "inWalletID",
			AccountID:    "inAccountID",
			Amount:       nums.PlainCurrencyNumber{
				Amount: "100.0",
				CurrencyType: "LTC",
			},
		},
		RequestStatus: OfferPending,
	}

	adapter := &chaincodetest.StubInterfaceAdapterMock{}

	adapter.On("GetState", testOperationID).Return(nil, nil)

	userJSON, err := json.Marshal(offer.OutUserInfo.toUserAccountInfo())
	assert.NoError(t, err)

	args := util.ToChaincodeArgs(invoke.VerifyAccountFunctionName, string(userJSON))
	adapter.On("InvokeChaincode", invoke.VerifyAccountChaincodeName, args, invoke.VerifyAccountChannelName).Return(shim.Success(nil))

	userJSON, err = json.Marshal(offer.InUserInfo.toUserAccountInfo())
	assert.NoError(t, err)

	args = util.ToChaincodeArgs(invoke.VerifyAccountFunctionName, string(userJSON))
	adapter.On("InvokeChaincode", invoke.VerifyAccountChaincodeName, args, invoke.VerifyAccountChannelName).Return(shim.Success(nil))

	offerJson, err := json.Marshal(offer)
	assert.NoError(t, err)

	adapter.On("PutState", testOperationID, offerJson).
		Run(func(args mock.Arguments) {
			// Assert that there was not change while offer data was processed
			// the offer itself
			jo := args.Get(1)
			assert.NotNil(t, jo)
			var request OfferRequest
			err := json.Unmarshal(jo.([]byte), &request)
			assert.NoError(t, err)
			assert.Equal(t, offer, request)
		}).
		Return(nil)

	err = mngr.AddOffer(offer, adapter)
	assert.NoError(t, err)
}

func TestOfferManagerImpl_Unmarshal(t *testing.T) {
	userJSON := "{\"operation_id\":\"OPERATION_ID\",\"out_info\":{\"user_id\":\"USER_ID\",\"wallet_id\":\"WALLET_ID\",\"account_id\":\"ACCOUNT_ID\",\"currency_type\":\"BTC\",\"amount\":1.0},\"in_info\":{\"user_id\":\"USER_ID\",\"wallet_id\":\"WALLET_ID\",\"account_id\":\"ACCOUNT_ID\",\"currency_type\":\"ETH\",\"amount\":2.0},\"request_status\":\"OfferPending\"}"

	var request *OfferRequest = &OfferRequest{}
	err := json.Unmarshal([]byte(userJSON), request)
	if err != nil {
		errMsg := fmt.Sprintf("failed to deserialize user offer request, due to error %s", err)
		logger.Error(errMsg)
	}

	assert.Equal(t, "OPERATION_ID", request.OperationID)
}
