package nums

import (
	"errors"
	"fmt"
	"github.com/shopspring/decimal"
	"github.com/sirupsen/logrus"
)

var logger = logrus.WithFields(logrus.Fields{
	"module": "nums",
})

//====================================================================================================================//
func (obj *CurrencyNumber) Add(b *CurrencyNumber) *CurrencyNumber {
	if !obj.AssetEquals(b) {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: errors.New(fmt.Sprintf("Assets are different: %s vs %s", obj.Asset.String(), b.Asset.String())),
		}
	}

	if obj.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: obj.Error,
		}
	}

	if b.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: b.Asset,
			Error: b.Error,
		}
	}

	res := decimal.NewFromFloat(0)
	res = res.Add(obj.Num)
	res = res.Add(b.Num)

	result := CurrencyNumber{
		Num:   res,
		Asset: correctAsset(obj, b),
		Error: nil,
	}

	return validate(&result)
}

func (obj *CurrencyNumber) AddSelf(b *CurrencyNumber) *CurrencyNumber {
	if !obj.AssetEquals(b) {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: errors.New(fmt.Sprintf("Assets are different: %s vs %s", obj.Asset.String(), b.Asset.String())),
		}
	}

	if obj.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: obj.Error,
		}
	}

	if b.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: b.Asset,
			Error: b.Error,
		}
	}
	obj.Num = obj.Num.Add(b.Num)
	obj.Asset = correctAsset(obj, b)

	return validate(obj)
}

func (obj *CurrencyNumber) Sub(b *CurrencyNumber) *CurrencyNumber {
	if !obj.AssetEquals(b) {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: errors.New(fmt.Sprintf("Assets are different: %s vs %s", obj.Asset.String(), b.Asset.String())),
		}
	}

	if obj.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: obj.Error,
		}
	}

	if b.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: b.Asset,
			Error: b.Error,
		}
	}
	res := decimal.NewFromFloat(0)
	res = res.Add(obj.Num)
	res = res.Sub(b.Num)

	result := CurrencyNumber{
		Num:   res,
		Asset: correctAsset(obj, b),
		Error: nil,
	}

	return validate(&result)
}

func (obj *CurrencyNumber) SubSelf(b *CurrencyNumber) *CurrencyNumber {
	if !obj.AssetEquals(b) {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: errors.New(fmt.Sprintf("Assets are different: %s vs %s", obj.Asset.String(), b.Asset.String())),
		}
	}

	if obj.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: obj.Error,
		}
	}

	if b.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: b.Asset,
			Error: b.Error,
		}
	}
	obj.Num = obj.Num.Sub(b.Num)
	obj.Asset = correctAsset(obj, b)

	return validate(obj)
}

func (obj *CurrencyNumber) Mul(b *CurrencyNumber) *CurrencyNumber {
	if !obj.AssetEquals(b) {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: errors.New(fmt.Sprintf("Assets are different: %s vs %s", obj.Asset.String(), b.Asset.String())),
		}
	}

	if obj.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: obj.Error,
		}
	}

	if b.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: b.Asset,
			Error: b.Error,
		}
	}

	res := decimal.NewFromFloat(0)
	res = res.Add(obj.Num)
	res = res.Mul(b.Num)

	result := CurrencyNumber{
		Num:   res,
		Asset: correctAsset(obj, b),
		Error: nil,
	}

	return validate(&result)
}

func (obj *CurrencyNumber) MulSelf(b *CurrencyNumber) *CurrencyNumber {
	if !obj.AssetEquals(b) {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: errors.New(fmt.Sprintf("Assets are different: %s vs %s", obj.Asset.String(), b.Asset.String())),
		}
	}

	if obj.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: obj.Error,
		}
	}

	if b.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: b.Asset,
			Error: b.Error,
		}
	}

	obj.Num = obj.Num.Mul(b.Num)
	obj.Asset = correctAsset(obj, b)

	return validate(obj)
}

func (obj *CurrencyNumber) Div(b *CurrencyNumber) *CurrencyNumber {
	if !obj.AssetEquals(b) {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: errors.New(fmt.Sprintf("Assets are different: %s vs %s", obj.Asset.String(), b.Asset.String())),
		}
	}

	if obj.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: obj.Error,
		}
	}

	if b.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: b.Asset,
			Error: b.Error,
		}
	}

	if b.Equals(Zero()) {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: errors.New("dividing by zero"),
		}
	}

	res := decimal.NewFromFloat(0)
	res = res.Add(obj.Num)
	res = res.Div(b.Num)

	result := CurrencyNumber{
		Num:   res,
		Asset: correctAsset(obj, b),
		Error: nil,
	}

	return validate(&result)
}

func (obj *CurrencyNumber) DivSelf(b *CurrencyNumber) *CurrencyNumber {
	if !obj.AssetEquals(b) {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: errors.New(fmt.Sprintf("Assets are different: %s vs %s", obj.Asset.String(), b.Asset.String())),
		}
	}

	if obj.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: obj.Asset,
			Error: obj.Error,
		}
	}

	if b.HasErrors() {
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: b.Asset,
			Error: b.Error,
		}
	}

	if b.Equals(Zero()) {
		obj.Error = errors.New("dividing by zero")
		return obj
	}

	obj.Num = obj.Num.Div(b.Num)
	obj.Asset = correctAsset(obj, b)

	return validate(obj)
}
