package txnmngr

import (
	"encoding/json"
	"fmt"

	"github.com/pkg/errors"
)

const (
	CommissionChannel     = "commissionChannel"
	CommissionChaincode   = "comissionCC"
	GetCommissionFunction = "getCommission"
)

type (
	// txnProcessType denote possible process types for transactions
	txnProcessType string

	// txnCommissionType
	txnCommissionType string

	// txnOperationType define operation types
	txnOperationType uint8
)

// Enumerate transactions process types
const (
	ExchangeWithCommission      txnProcessType = "ExchangeWithCommission"
	TransferSenderCommissions   txnProcessType = "TransferSenderCommissions"
	TransferReceiverCommissions txnProcessType = "TransferReceiverCommissions"
	DepositFunds                txnProcessType = "DepositFunds"

	SenderCommission   txnCommissionType = "SenderCommission"
	ReceiverCommission txnCommissionType = "ReceiverCommission"

	LockAmount                 txnOperationType = iota
	LockCommission
	TransferAmountToUser
	TransferCommissionToSystem
)

// String converts commission type into its string representation
func (c txnCommissionType) String() string {
	return string(c)
}

// processTypes maps all possible values of txnProcessType
var processTypes = map[txnProcessType]int8{
	ExchangeWithCommission:      0,
	TransferSenderCommissions:   1,
	TransferReceiverCommissions: 2,
	DepositFunds:                3,
}

// validate check correctness of the txnProcessType
// only defined values are allowed
func (t txnProcessType) validate() bool {
	_, ok := processTypes[t]
	return ok
}

// String returns string representation for
func (t txnOperationType) String() string {
	return map[txnOperationType]string{
		LockAmount:                 "LockAmount",
		LockCommission:             "LockCommission",
		TransferAmountToUser:       "TransferAmountToUser",
		TransferCommissionToSystem: "TransferCommissionToSystem",
	}[t]
}

// AccountInfo account information with wallet id and account id
type AccountInfo struct {
	WalletID  string `json:"wallet_id,omitempty"`
	AccountID string `json:"account_id,omitempty"`
}

// InKeyForCurrency generate key for "in" direction based on account information
// given currency id
func (info AccountInfo) InKeyForCurrency(currencyType string) string {
	return fmt.Sprintf("in.%s.%s", currencyType, info.AccountID)
}

// OutKeyForCurrency generate key for "out" direction based on account information
// given currency id
func (info AccountInfo) OutKeyForCurrency(currencyType string) string {
	return fmt.Sprintf("out.%s.%s", currencyType, info.AccountID)
}

// InKeyForCurrency generate key for "in" direction based on account information
// given currency id
func (info CryptocurrencyAccountInfo) InKeyForCurrency(currencyType string) string {
	return fmt.Sprintf("in.%s.%s", currencyType, info.AccountID)
}

// OutKeyForCurrency generate key for "out" direction based on account information
// given currency id
func (info CryptocurrencyAccountInfo) OutKeyForCurrency(currencyType string) string {
	return fmt.Sprintf("out.%s.%s", currencyType, info.AccountID)
}

// AccountInfo cpatures account info,
// including accountID, userID and walletID
type UserAccountInfo struct {
	UserID string `json:"user_id,omitempty"`
	// Embed account information into struct
	// with walletID and accountID
	AccountInfo
}

// String returns string representation of the account information
func (info UserAccountInfo) String() string {
	return fmt.Sprintf("account information: user id <%s>, wallet id <%s>, account id <%s>",
		info.UserID, info.WalletID, info.AccountID)
}

// TxnFundsLockInput encapsulates input parameters
// for lock funds transaction type
type TxnFundsLockInput struct {
	OperationID       string            `json:"operation_id"`
	CurrencyType      string            `json:"currency_type"`
	FromUser          UserAccountInfo   `json:"from_user"`
	ToUser            UserAccountInfo   `json:"to_user"`
	LockAccount       AccountInfo       `json:"lock_account"`
	CommissionAccount AccountInfo       `json:"commission_account"`
	CommissionType    txnCommissionType `json:"commission_type"`
	CustomerType      string            `json:"customer_type"`
	PartnerCode       string            `json:"partner_code"`
	ProcessType       txnProcessType    `json:"process_type"`
	Amount            float32           `json:"amount"`
}

//FundsTransferInput
type FundsTransferInput struct {
	Txns []TxnFundsLockInput `json:"transactions"`
}

// TransactionRecord transaction struct which denotes transaction
// fields which are necessary for processing of
// transactions and will be used to present transaction in the ledger
type TransactionRecord struct {
	Key           string           `json:"key"`
	OperationID   string           `json:"operation_id,omitempty"`
	ProcessType   txnProcessType   `json:"process_type,omitempty"`
	OperationType txnOperationType `json:"operation_type,omitempty"`
	User          UserAccountInfo  `json:"user,omitempty"`
	CurrencyType  string           `json:"currency_type,omitempty"`
	Amount        float32          `json:"amount,omitempty"`
	HashCode      string           `json:"hash_code,omitempty"`
	DependencyKey string           `json:"dependency_key,omitempty"` // transaction dependency in a form of "txnType.currencyType.accountId"
	Timestamp     int64            `json:"timestamp"`
}

func (tx *TransactionRecord) Save(adapter chaincodeStubInterfaceAdapter) error {
	// Serialize transaction record into JSON
	jsonRecordLockFunds, err := json.Marshal(tx)
	if err != nil {
		errMsg := fmt.Sprintf("failed to marhsal record data, operation id %s, error %s", tx.OperationID, err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// Generate key based on transaction id
	recKey := fmt.Sprintf("%s_%s_%d", adapter.GetTxID(), tx.User.UserID, tx.OperationType)
	err = adapter.PutState(recKey, jsonRecordLockFunds)
	if err != nil {
		errMsg := fmt.Sprintf("failed to add new transaction (%s), operation id %s, error %s", tx, tx.OperationID, err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	return nil
}

// AccountBalance denote balance of the
// account for given currency id
type AccountBalance struct {
	AccountID    string  `json:"account_id,omitempty"`
	CurrencyType string  `json:"currency_type,omitempty"`
	Balance      float32 `json:"balance"`
}

// queryKey document key
type queryKey struct {
	Key           string `json:"key"`
	OperationID   string `json:"operation_id,omitempty"`
	DependencyKey string `json:"dependency_key,omitempty"`
}

// querySelector document query selector
type querySelector struct {
	Selector queryKey `json:"selector"`
}

// selectTx use selector to query for transaction record from the state
func (q querySelector) selectTx(adapter chaincodeStubInterfaceAdapter) (*TransactionRecord, error) {
	lockedTxQuery, err := json.Marshal(q)
	if err != nil {
		errMsg := fmt.Sprintf("failed to serialize query to find locked transaction, due to %s", err)
		logger.Error(errMsg)
		return nil, err
	}

	// Locked funds transaction record
	var lockedFundsRecord TransactionRecord
	it, err := adapter.GetQueryResult(string(lockedTxQuery))
	if err != nil {
		errMsg := fmt.Sprintf("failed to retreive transaction with locked funds, query %s, due to %s", lockedTxQuery, err)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	if !it.HasNext() {
		errMsg := fmt.Sprintf("locked funds was not found, out key %s", q.Selector.Key)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	lockedFundsJsonRecord, err := it.Next()
	if err != nil {
		errMsg := fmt.Sprintf("filed to read next record from iterator to get transaction record for operation id %s, due to %s", q.Selector.OperationID, err)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}
	// Get locked funds information from the state db
	err = json.Unmarshal(lockedFundsJsonRecord.Value, &lockedFundsRecord)
	if err != nil {
		errMsg := fmt.Sprintf("failed to deserialize transaction record for operation id %s, due to %s", q.Selector.OperationID, err)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	return &lockedFundsRecord, nil
}
