package txnmngr

import (
	"encoding/json"
	"errors"
	"fmt"

	"github.com/hyperledger/fabric/core/chaincode/shim"
	pb "github.com/hyperledger/fabric/protos/peer"
	"github.com/hyperledger/fabric/common/util"
)

// managerAux
type managerAux struct {
	chaincodeStubInterfaceAdapter
	timeGetterFunc func() (int64, error)
}

func (m *managerAux) GetTime() (int64, error) {
	return m.timeGetterFunc()
}

// transactionChaincodeAction encapsulates transaction chaincode operation
type transactionChaincodeAction func(args []string, adapter *managerAux) pb.Response

const (
	LockFundsOnAccount     = "lockFundsOnAccount"
	AccountCurrencyBalance = "accountCurrencyBalance"
	TransferFunds          = "transferFunds"
	LoadCryptocurrency     = "loadCryptocurrency"

	CheckAccountChaincode = "accountCC"
	CheckAccountFunction  = "checkAndLoadAccount"
	CheckAccountChannel   = "accountchannel"

	CryptocurrenciesChaincode   = "cryptocurrencyCC"
	GetCryptocurrenciesFunction = "getCryptocurrencies"
	CryptocurrencyChannel       = "cryptocurrencychannel"
)

// transactionChaincode transaction chaincode to operate on transactions
// lock funds, lock commissions, transfer money and etc...
type transactionChaincode struct {
	actions map[string]transactionChaincodeAction
}

//Init initialize chaincode bootstrap state
func (t *transactionChaincode) Init(stub shim.ChaincodeStubInterface) pb.Response {
	logger.Info("Start initializing chaincode...")
	return shim.Success([]byte("Chaincode initialized successfully"))

}

//Invoke chaincode logic based on input parameters, first parameter has to be function name
func (t *transactionChaincode) Invoke(stub shim.ChaincodeStubInterface) pb.Response {
	function, args := stub.GetFunctionAndParameters()
	logger.Info("Invoke function " + function)

	if action, ok := t.actions[function]; ok {
		return action(args, &managerAux{
			chaincodeStubInterfaceAdapter: stub,
			timeGetterFunc: func(stub shim.ChaincodeStubInterface) func() (int64, error) {
				return func() (int64, error) {
					tt, err := stub.GetTxTimestamp()
					if err != nil {
						return 0, err
					}
					return tt.Seconds, nil
				}
			}(stub),
		})
	}

	logger.Errorf("Invalid invoke function name, received %s.", function)
	return shim.Error("Invalid invoke function name.")
}

type CryptocurrencyTypeInput struct {
	CurrencyTypeList []string `json:"currency_type_list"`
}

// NewTransactionChaincode crate a new instance of transaction chaincode
func NewTransactionChaincode() shim.Chaincode {
	checkAccountFunction := func(user *UserAccountInfo, adapter chaincodeStubInterfaceAdapter) (*UserAccountInfo, error) {
		// load user parameters into accountCheck chaincode
		userJSON, err := json.Marshal(user)
		if err != nil {
			errMsg := fmt.Sprintf("Cannot serialize user account info object, validation failed")
			logger.Error(errMsg)
			return nil, errors.New(errMsg)
		}
		// Invoke account check chaincode
		resp := adapter.InvokeChaincode(CheckAccountChaincode, util.ToChaincodeArgs(CheckAccountFunction, string(userJSON)), CheckAccountChannel)
		// We got error status
		if resp.Status != int32(shim.OK) {
			logger.Errorf("verification of user id %s failed, with following response %s", user.UserID, resp.Message)
			return nil, errors.New(resp.Message)
		}
		// No error, account verified
		var userAccount UserAccountInfo
		err = json.Unmarshal(resp.Payload, &userAccount)
		if err != nil {
			errMsg := fmt.Sprintf("Failed to deserialize user account info, due to %s", err)
			logger.Error(errMsg)
			return nil, err
		}
		return user, nil
	}

	getCryptocurrenciesFunction := func(cryptocurrencyTypeInput CryptocurrencyTypeInput, adapter chaincodeStubInterfaceAdapter) (*CryptocurrencyStoreOutput, error) {
		cryptocurrencyTypeInputJSON, err := json.Marshal(cryptocurrencyTypeInput)
		if err != nil {
			errMsg := fmt.Sprintf("Cannot serialize user account info object, validation failed")
			logger.Error(errMsg)
			return nil, errors.New(errMsg)
		}
		// Invoke account check chaincode

		resp := adapter.InvokeChaincode(CryptocurrenciesChaincode, util.ToChaincodeArgs(GetCryptocurrenciesFunction, string(cryptocurrencyTypeInputJSON)), CryptocurrencyChannel)
		// We got error status
		if resp.Status != int32(shim.OK) {
			logger.Errorf("Failed to invoke GetCryptocurrencies from CryptocurrenciesChaincode, with following error message %s", resp.Message)
			return nil, errors.New(resp.Message)
		}
		// No error, account verified
		var cryptocurrencyStoreOutput CryptocurrencyStoreOutput
		err = json.Unmarshal(resp.Payload, &cryptocurrencyStoreOutput)
		if err != nil {
			errMsg := fmt.Sprintf("Failed to deserialize user CryptocurrencyStoreOutput, due to %s", err)
			logger.Error(errMsg)
			return nil, err
		}
		return &cryptocurrencyStoreOutput, nil
	}

	mngr := NewTransactionManager(checkAccountFunction, getCryptocurrenciesFunction)

	return &transactionChaincode{
		actions: map[string]transactionChaincodeAction{
			LockFundsOnAccount: func(args []string, adapter *managerAux) pb.Response {
				if len(args) != 1 {
					errMsg := fmt.Sprintf("LockFundsOnAccount: Invalid number of arguments expecting 1, while got %d", len(args))
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}
				var input TxnFundsLockInput
				err := json.Unmarshal([]byte(args[0]), &input)
				if err != nil {
					errMsg := fmt.Sprintf("Invalid transaction input format, failed due to %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}
				// Trying to lock funds on account
				err = mngr.LockFundsOnAccount(&input, adapter)
				if err != nil {
					errMsg := fmt.Sprintf("Failed to lock funds on account %s, error %s", input.FromUser.AccountID, err)
					logger.Errorf(errMsg)
					return shim.Error(errMsg)
				}

				return shim.Success(nil)
			},
			AccountCurrencyBalance: func(args []string, adapter *managerAux) pb.Response {
				if len(args) != 2 {
					errMsg := fmt.Sprintf("AccountCurrencyBalance: Invalid number of arguments expecting 2, while got %d", len(args))
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}
				accountID := args[0]
				currencyType := args[1]

				// Use transaction manager to get account balance with given currency
				balance, err := mngr.AccountCurrencyBalance(accountID, currencyType, adapter)
				if err != nil {
					errMsg := fmt.Sprintf("Failed to obtain total balance for account id %s, currency id %s, due to %s.", accountID, currencyType, err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}
				jsonResult, err := json.Marshal(&AccountBalance{
					AccountID:    accountID,
					CurrencyType: currencyType,
					Balance:      balance,
				})
				if err != nil {
					errMsg := fmt.Sprintf("Failed to serialize account id %d balance responce, currency id %s, due to %s", accountID, currencyType, err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}
				return shim.Success(jsonResult)
			},
			TransferFunds: func(args []string, adapter *managerAux) pb.Response {
				if len(args) != 1 {
					errMsg := fmt.Sprintf("TransferFunds: Invalid number of arguments expecting 1, while got %d", len(args))
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}
				var txns FundsTransferInput
				err := json.Unmarshal([]byte(args[0]), &txns)
				if err != nil {
					errMsg := fmt.Sprintf("Invalid transaction input format, failed due to %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}
				for _, txn := range txns.Txns {
					err := mngr.TransferFunds(&txn, adapter)
					if err != nil {
						errMsg := fmt.Sprintf("Failed to transfer funds, due to %s", err)
						logger.Error(errMsg)
						return shim.Error(errMsg)
					}
				}

				return shim.Success(nil)
			},
			LoadCryptocurrency: func(args []string, adapter *managerAux) pb.Response {
				if len(args) == 1 {
					errMsg := fmt.Sprintf("LoadCryptocurrency: Invalid number of arguments expecting 0, while got %d", len(args))
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				var cryptocurrencyTypeInput CryptocurrencyTypeInput
				cryptocurrencyTypeByteArray := []byte(args[0])
				err := json.Unmarshal(cryptocurrencyTypeByteArray, &cryptocurrencyTypeInput)
				if err != nil {
					errMsg := fmt.Sprintf("Invalid CryptocurrencyTypeInput input format, failed due to %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				importCryptocurrencyResult, err := mngr.LoadCryptocurrency(cryptocurrencyTypeInput, adapter)
				if err != nil {
					errMsg := fmt.Sprintf("Failed to Load Cryptocurrency, due to %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				bytes, err := json.Marshal(importCryptocurrencyResult)
				if err != nil {
					errMsg := fmt.Sprintf("Failed to Marshal funds, due to %s", err)
					logger.Error(errMsg)
					return shim.Error(err.Error())
				}

				return shim.Success(bytes)
			},
		},
	}
}
